﻿; ------------------------------------------------------------------------------------
;- Internal
; ------------------------------------------------------------------------------------

Prototype SC_EventCallback(*SC, EventType, Size.q, curSize.q, CacheAccess)

; Ab dieser Anzahl von Einträgen wird die HashTable zum Suchen genutz (Defaultwert)
#SC_HT_Limit = 200
; Ab wie vielen Zugriffen landet ein Eintrag im unpacked - Cache?
#SC_UD_Limit = 2
; Version
#SC_Version = 104

; Includes
XIncludeFile "common\FastMem.pbi" ; Schnelle Memoryfunktionen, überschreibt CopyMemory()
XIncludeFile "common\LinkedList.pbi" ; Dynamische Linked List
XIncludeFile "common\packer.pbi" ; Universelle Kompressionsbibliothek
XIncludeFile "common\HashTable.pbi" ; Hashtable um die Zugriffszeiten bei vielen Einträgen zu Optimieren
XIncludeFile "common\GrowableMemory.pbi" ; Wachsender Speicher
XIncludeFile "common\Logger.pbi" ; Logsystem
XIncludeFile "FreeSpaceController.pbi" ; Speicherplatzverwaltung
XIncludeFile "JobQueue_MultiThread.pbi" ; Framework zur parallelisierung
XIncludeFile "ECF.pbi" ; EFC - Dateiverwaltung

EnableExplicit

Structure SC_Entry_XtraData_Entry_Dynamics ; Dynamische Einträge für die XtraDaten eines Entrys
  *Name ; Name des Entrys
  Value.q ; Wert
EndStructure

Structure SC_Entry_XtraData_Entry ; XtraDaten eines Entrys
  Type.a ; Typ (String/Quad)
  NameSize.l ; Name
  StringSize.l ; Größe des Strings, falls Type = String
  Dyn.SC_Entry_XtraData_Entry_Dynamics ; Dynamische Daten
EndStructure

Structure SC_Entry_Dynamics
  *Name ; Name
  *XtraList.LL_Data ; Liste für Extra - Einträge
  tId.q ; Temporäre Eintrags - Id
  tAccessCount.i ; Aktueller Zugriffszähler
  AsyncInProgress.a ; Eintrag wird gerade Asyncron verarbeitet
  AsyncMode.a ; AsyncMode: Read oder Write? 
EndStructure

; Structures
Structure SC_Entry
  OwnSize.w ; Größe der Struktur: um bei Änderungen der Struktur abwärts-kompatibel zu bleiben.
  Position.q ; Position des Eintrags in der Datei
  Size.q ; Unkomprimierte größe des Eintrags
  PackedSize.q ; Gepackte Größe des Eintrags
  Compressed.a ; sind die Daten komprimiert
  NameSize.l ; Größe des Namens
  XtraDataCount.l ; Anzahl der Extra - Einträge
  Packer.a ; Welcher Packer wurde verwendet?
  ; <- Neue Werte hier einfügen
  Dyn.SC_Entry_Dynamics ; MUSS!!!! Am Ende stehen! enthält Dynamische Daten
EndStructure

Structure SC_Current ; Aktuelles Entry
  *Entry.SC_Entry
  *Mem
EndStructure

Structure SC_Data
  EnumLoadMem.i ; Speicher bei Enumeration laden?
  *Entrys.LL_Data ; Einträge - Liste
  EntryHT.i ; Einträge - Hashtable
  HTLimit.i ; Ab wie vielen Einträgen soll die Hashtable verwendet werden?
  UDLimit.i ; Ab wie vielen Zugriffen landet ein Eintrag im Unpacked - Cache?
  FSC.i ; FreeSpaceController
  ECF.i ; ECF - Dateiverwaltung
  UseUDCache.a ; Wird der Cache verwendet?
  UDCache.i ; Cache für entpackte Daten
  Logger.i ; Logger
  tIdIndex.q ; Aktuelle tId
  
  EnableTransaction.a ; Transaction aktivieren
  EnableAscnyMode.a ; Asyncrones packen/schreiben aktivieren
  ContainerChanged.a ; Container wurde verändert
  
  JQ.i ; Jobqueue zur parallelisierung
  AsyncLastAccessJob.i ; Letzter SchreibJob
  AsyncAccessMutex.i  ; Zugriffsmutex
  AsyncCopyMemory.a ; Asyncspeicher kopieren?
  
  CPos.q ; aktuelle Position in der Datei
  Current.SC_Current ; Aktueller Eintrag
  *Event.SC_EventCallback ; Event - Callback
EndStructure

Structure SC_Head
  Magic.l ; Magic (ist das eine ECF - Datei?)
  OffSet.q ; Offset des Indexes 
  EntryCount.q ; Anzahl Einträge
  FSCSize.q ; FreeSpace Size
EndStructure

Structure SC_Async_Launcher
  *SC.SC_Data
  *SCE.SC_Entry
  PackerInfo.Packer_Info
  *Mem
  *PackedMem
  MemSize.i
  CompressJob.i
  ParentWriter.i
EndStructure

; Enumerations
Enumeration
  #SC_DefaultBlockSize = #ECF_DefaultBlockSize
  #SC_DefaultCacheSize = #ECF_DefaultCacheSize
  #SC_Magic = $23651290
  #SC_Async_Write = 0
  #SC_Async_Read = 1
EndEnumeration

Enumeration
  #SC_Event_Read = #ECF_Access_Read
  #SC_Event_Write = #ECF_Access_Write
  #SC_Event_ReBlock = #ECF_Access_ReBlock
  #SC_Event_CompressStart
  #SC_Event_CompressFinished
EndEnumeration

Enumeration
  #SC_zLib = #Packer_zLib
  #SC_LZMA = #Packer_LZMA
  #SC_bZip2 = #Packer_bZip2
  #SC_aPLib = #Packer_aPLib
  #SC_Uncompressed = #Packer_Uncompressed
EndEnumeration

; Laden eines Eintrags, Leseteil
Macro _SC_LoadEntry_Read(_SC)
  If _SC#\UseUDCache ; Suchen wir mal im Cache ...
    If Cache_isCached(_SC#\UDCache, _SC#\Current\Entry\Dyn\tId)
      ; Eintrag vorhanden, einfach aus dem Cache laden, spart Entschlüsseln & Entpacken :D
      _SC#\Current\Mem = AllocateMemory(_SC#\Current\Entry\Size)
      CopyMemory(Cache_GetMem(_SC#\UDCache), _SC#\Current\Mem, _SC#\Current\Entry\Size)
      isUDCached = #True
    EndIf
  EndIf
  
  If Not isUDCached ; Wir sind nicht im Cache, normal laden
    *pMem = AllocateMemory(_SC#\Current\Entry\PackedSize)
    ECF_Seek(_SC#\ECF, _SC#\Current\Entry\Position)
    ECF_ReadData(_SC#\ECF, *pMem, _SC#\Current\Entry\PackedSize)
  EndIf
EndMacro

; Laden eines Entrys, Entpackteil
Macro _SC_LoadEntry_UnCompress(_SC)
  ; Eintrag entpacke wenn nötig
  If _SC#\Current\Entry\Compressed
    _SC#\Current\Mem = Packer_Uncompress(*pMem, _SC#\Current\Entry\PackedSize, _SC#\Current\Entry\Size, _SC#\Current\Entry\Packer)
    FreeMemory(*pMem)
  Else
    _SC#\Current\Mem = *pMem
  EndIf
    
  ; Daten landen im Unpacked - Cache
  If _SC#\UDLimit <= _SC#\Current\Entry\Dyn\tAccessCount ; Zugriffslimit für UnpackedCache wurde erreicht!
    Cache_AddEntry(_SC#\UDCache, _SC#\Current\Entry\Dyn\tId, _SC#\Current\Mem, _SC#\Current\Entry\Size)
    ; Lese/Schreibe Cache Fluschen, nun unnötig.
    ECF_Seek(_SC#\ECF, _SC#\Current\Entry\Position)
    ECF_ClearCache(_SC#\ECF, _SC#\Current\Entry\PackedSize)
  EndIf
EndMacro

; Enpackt den Eintrag aus *SCE nach \Current\Mem
Macro _SC_LoadEntry(_SC)
  ; Eintrag laden
  Protected *pMem, isUDCached
  
  _SC_LoadEntry_Read(_SC)
  
  If Not isUDCached
    _SC_LoadEntry_UnCompress(_SC)
  EndIf
EndMacro

; Läd ein Entry komplett Incl. lesen, entpacken, sync+async
Macro _SC_LoadEntry_All(_SC, _SCE, _LoadMem, _WaitForAsync = #False)
  Protected *Async.SC_Async_Launcher
  
  ; Wird gerade ein Entry Asnc geladen kann kein anderes Entry selektiert werden!
  If _SC#\Current\Entry
    If _SC#\Current\Entry\Dyn\AsyncInProgress And (_SC#\Current\Entry\Dyn\AsyncMode = #SC_Async_Read)
      CompilerIf _WaitForAsync ; soll auf den Async-Prozess gewartet werden?
        While _SC#\Current\Entry\Dyn\AsyncInProgress ; Warten bis alles fertig ist
          Delay(1)
        Wend
      CompilerElse ; Beenden
        ProcedureReturn #False 
      CompilerEndIf
    EndIf
  EndIf
      
  If _SC#\Current\Mem ; Aktuellen Speicher freigeben
    FreeMemory(_SC#\Current\Mem)
  EndIf
  _SC#\Current\Entry = _SCE# ; Eintrag aktuell setzen
  _SCE#\Dyn\TAccessCount + 1 ; Zugriffszähler erhöhen
      
  If _LoadMem ; soll auch der Speicher geladen werden?
    If _SC#\EnableAscnyMode
      If _SCE#\Dyn\AsyncInProgress ; Daten werden erst geschrieben!
        CompilerIf _WaitForAsync ; soll gewartet werden?
          While _SCE#\Dyn\AsyncInProgress ; Warten bis alles fertig ist
            Delay(1)
          Wend
        CompilerElse
          Logger_Log(_SC#\Logger, "entry is not ready, async write in progress!", #Logger_Warning)
          _SC#\Current\Entry = #Null
          RtVar = #False
        CompilerEndIf
      Else ; alles Ok, Async launchen
        _SCE#\Dyn\AsyncInProgress = #True
        _SCE#\Dyn\AsyncMode = #SC_Async_Read
           
        ; Async Launcher einstellen
        *Async = AllocateMemory(SizeOf(SC_Async_Launcher))
        *Async\SC = _SC
        *Async\SCE = _SCE
            
        ; Job launchen
        LockMutex(_SC#\AsyncAccessMutex) ; Filezugriffe müssen gelenkt ablaufen
        *Async\ParentWriter = _SC#\AsyncLastAccessJob ; Parentjob merken
        _SC#\AsyncLastAccessJob = JQ_Add(_SC#\JQ, @_SC_Async_Reader(), *Async, #Job_HighestPriority, _SC#\AsyncLastAccessJob)
        UnlockMutex(_SC#\AsyncAccessMutex)
            
        RtVar = #True
      EndIf
    Else ; "Normales" Laden
      _SC_LoadEntry(_SC)
      ; Und Fertig :D
      If _SC#\Current\Mem
        RtVar = #True
      EndIf
    EndIf
  Else
    _SC#\Current\Mem = #Null
    RtVar = #True
  EndIf  
EndMacro

Macro _SC_CompressData(_SC, _SCE, _PackerInfo, _Mem, _MemSize, _PackedMem, _PackedMemSize)
  ; Event absetzen
  _SC_SendEvent(_SC, #SC_Event_CompressStart, _Mem, _MemSize, #Null)
    
  ; Komprimierung der Daten
  _PackedMem = Packer_Compress(_Mem, _MemSize, @_PackerInfo)
  If _PackerInfo#\DataOk ; Kompression Ok?
    _PackedMemSize = _PackerInfo#\MemSize  
  Else
    _PackedMem = _Mem
    _PackedMemSize = _MemSize
  EndIf
    
  ; CompressDaten schreiben
  _SCE#\PackedSize = _PackedMemSize ; Gepackte größe
  _SCE#\Compressed = _PackerInfo#\DataOk ; wurde korrekt gepackt?
    
  ; Event absetzen
  _SC_SendEvent(_SC, #SC_Event_CompressFinished, _MemSize, _PackedMemSize, #Null)
EndMacro

Macro _SC_WriteData(_SC, _SCE, _Mem)
  ; Sucht einen passenden Platz für die Daten
  Protected FSCfound
  
  _SCE#\Position = FSC_Find(_SC#\FSC, _SCE#\PackedSize, @FSCfound)
  If Not FSCfound ; Gibt Nix freies in der Datei, Eintrag muss hinten dran
    _SCE#\Position = _SC#\CPos 
  EndIf
    
  ECF_Seek(_SC#\ECF, _SCE#\Position)
  ECF_WriteData(_SC#\ECF, _Mem, _SCE#\PackedSize)
    
  If Not FSCfound ; Zeiger weiterstellen wenn nötig
    _SC#\CPos + _SCE#\PackedSize
  EndIf
    
  If _SCE#\Compressed ; Speicher freigeben
    FreeMemory(_Mem)
  EndIf  
EndMacro

Macro _SC_DeleteEntry(_SC, _SCE, _SafeDelete = #False, _ClearOnly = #False, _GMem = #False)
  Protected *SCEXE.SC_Entry_XtraData_Entry
  
  ; Der zu löschende Eintrag ist der aktuelle Eintrag
  
  CompilerIf _ClearOnly = #False
    ; Zu löschender Eintrag ist Current!
    If _SC#\Current\Entry = _SCE
      If _SC#\Current\Mem
        FreeMemory(_SC#\Current\Mem)
        _SC#\Current\Mem = #Null
      EndIf
       _SC#\Current\Entry = #Null
    EndIf
    
    ; Freigegebenen Speicherplatz in den Controller speichern
    FSC_Add(_SC#\FSC, _SCE#\Position, _SCE#\PackedSize) 
    
    ; Eintrag aus Hashtable löschen
    HT_SetValue(_SC#\EntryHT, PeekS(_SCE#\Dyn\Name), #Null)
  CompilerEndIf  
  
  ; Eintrag löschen
  LL_Delete(_SC#\Entrys)
  
  If _GMem ; Daten schreiben
    If _SCE#\Dyn\XtraList
      _SCE#\XtraDataCount = LL_Count(_SCE#\Dyn\XtraList)
    EndIf
    GMem_WriteData(_GMem, _SCE, _SCE#\OwnSize)
    GMem_WriteData(_GMem, _SCE#\Dyn\Name, _SCE#\NameSize)
  EndIf  
  
  ; Name freigeben
  FreeMemory(_SCE#\Dyn\Name)
      
  ; Xtra-Einträge löschen wenn da
  If _SCE#\Dyn\XtraList
    LL_Reset(_SCE#\Dyn\XtraList)
    While LL_Next(_SCE#\Dyn\XtraList) ; Alle ExtraDaten durchgehen
      *SCEXE = LL_Get(_SCE#\Dyn\XtraList)
      LL_Delete(*SCE\Dyn\XtraList)
      
      If _GMem ; Infos in die Datei schreiben
        ; Allgemeine Infos + Namen speichern
        GMem_WriteData(_GMem, *SCEXE, SizeOf(SC_Entry_XtraData_Entry)-SizeOf(SC_Entry_XtraData_Entry_Dynamics))
        GMem_WriteData(_GMem, *SCEXE\Dyn\Name, *SCEXE\NameSize)  
      EndIf
      
      FreeMemory(*SCEXE\Dyn\Name) ; Namenspeicher freigeben
      
      If *SCEXE\Type = #PB_String
        If _GMem  ; Infos in die Datei schreiben
          GMem_WriteData(_GMem, *SCEXE\Dyn\Value, *SCEXE\StringSize)  
        EndIf
        FreeMemory(*SCEXE\Dyn\Value)
      ElseIf _GMem  ; Infos in die Datei schreiben
        GMem_WriteData(_GMem, @*SCEXE\Dyn\Value, SizeOf(Quad))
      EndIf
      FreeMemory(*SCEXE)
    Wend
    LL_Free(_SCE#\Dyn\XtraList)
  EndIf
  
  CompilerIf _ClearOnly = #False; nicht nötig wenn wir nur clearen
    If _SafeDelete ; Sicheres löschen (überschreiben mit Zufallswerten)
      ECF_Seek(_SC#\ECF, _SCE#\Position)
      ECF_SafeDelete(_SC#\ECF, _SCE#\PackedSize)
    EndIf
      
    ; Wenn der Delayed Write aktiv ist nicht löschen, sonst kontraproduktiv
    If Not _SC#\EnableTransaction
      ECF_Seek(_SC#\ECF, _SCE#\Position)
      ECF_ClearCache(_SC#\ECF, _SCE#\PackedSize)
    EndIf
      
    If _SC#\UseUDCache ; Aus dem Unpacked Cache löschen
      Cache_DeleteEntry(_SC#\UDCache, _SCE#\Dyn\tId)
    EndIf
  CompilerEndIf
  
  FreeMemory(_SCE)  
EndMacro

; Sucht nach einem Eintrag und gibt den *SCE Eintrag darauf zurück falls gefunden
Procedure _SC_FindEntry(*SC.SC_Data, *Name, _EnableHTSearch = #True)
  Protected *SCE.SC_Entry
  
  With *SC
    If (LL_Count(\Entrys) >= \HTLimit) And _EnableHTSearch ; Hashtablesuche bei vielen Einträgen verwenden, bessere Performance
      ProcedureReturn HT_GetValue(\EntryHT, PeekS(*Name))  
    Else ; Sonst alle Listenelemente durchgehen
      LL_Reset(\Entrys)
      While LL_Next(\Entrys)
        *SCE = LL_Get(\Entrys)
        If (CompareMemoryString(*SCE\Dyn\Name, *Name) = #PB_String_Equal) ; Treffer
          ProcedureReturn *SCE
        EndIf
      Wend
    EndIf
  EndWith
EndProcedure

; XtraData - Eintrag finden
Procedure _SC_FindXtraData(*SC.SC_Data, *Name)
  Protected *SCEXE.SC_Entry_XtraData_Entry
  
  ; Liste durchgehen
  With *SC
    If \Current\Entry And \Current\Entry\Dyn\XtraList
      LL_Reset(\Current\Entry\Dyn\XtraList)
      While LL_Next(\Current\Entry\Dyn\XtraList)
        *SCEXE = LL_Get(\Current\Entry\Dyn\XtraList)
        If (CompareMemoryString(*SCEXE\Dyn\Name, *Name) = #PB_String_Equal) ; Treffer
          ProcedureReturn *SCEXE
        EndIf
      Wend
    EndIf
  EndWith
EndProcedure

; Extradaten zum Eintrag speichern
Procedure _SC_SetXtraData(*SC.SC_Data, Name$, Type.a, XData.q)
  Protected *SCEXE.SC_Entry_XtraData_Entry, AddToList = #True
  
  With *SC
    If \Current\Entry
      If Not \Current\Entry\Dyn\XtraList ; Liste anlegen wenn nötig
        \Current\Entry\Dyn\XtraList = LL_New()
      EndIf
      
      *SCEXE = _SC_FindXtraData(*SC, @Name$)
      
      ; Eintrag finden oder Erstellen wenn nicht vorhanden
      If *SCEXE
        AddToList = #False
        If *SCEXE\Type = #PB_String
          FreeMemory(*SCEXE\Dyn\Value)
          *SCEXE\StringSize = #Null
        EndIf
      Else
        ; Eintrag erstellen und Namen schreiben
        *SCEXE = AllocateMemory(SizeOf(SC_Entry_XtraData_Entry))
        *SCEXE\NameSize = StringByteLength(Name$) + SizeOf(Character)
        *SCEXE\Dyn\Name = AllocateMemory(*SCEXE\NameSize)
        PokeS(*SCEXE\Dyn\Name, Name$)
      EndIf
      
      ; Inhalt schreiben
      *SCEXE\Type = Type
      Select *SCEXE\Type
        Case #PB_String
          *SCEXE\StringSize = MemoryStringLength(XData) + SizeOf(Character)
          *SCEXE\Dyn\Value = AllocateMemory(*SCEXE\StringSize)
          CopyMemory(XData, *SCEXE\Dyn\Value, *SCEXE\StringSize)
        Default
          *SCEXE\Dyn\Value = XData
      EndSelect
      
      If AddToList ; zu ExtraDaten hinzufügen
        LL_Add(\Current\Entry\Dyn\XtraList) : LL_Set(\Current\Entry\Dyn\XtraList, *SCEXE)
      EndIf
      
      ; Container wurde verändert
      \ContainerChanged = #True
      
      ProcedureReturn #True
    EndIf
  EndWith  
EndProcedure

; Setzt ein Event ab.
Macro _SC_SendEvent(_SC, _EventType, _Size, _curSize, _CacheAccess)
  If _SC#\Event 
    _SC#\Event(_SC, _EventType, _Size, _curSize, _CacheAccess)
  EndIf
EndMacro

; Wrapper für das ECF - Callback
Procedure _SC_ECFFileAccessWrapper(ECF_File, AccessType, Size.q, curSize.q, CacheAccess, *SC.SC_Data)
  With *SC
    _SC_SendEvent(*SC, AccessType, Size.q, curSize.q, CacheAccess)
  EndWith
EndProcedure

; Uncompress procedure für Asyncrones lesen
Procedure _SC_Async_UnCompressor(*Async.SC_Async_Launcher)
  Protected *pMem
  
  With *Async
    *pMem = *Async\Mem
    _SC_LoadEntry_UnCompress(\SC) ; Entpacken
    \SCE\Dyn\AsyncInProgress = #False ; Async Fertig
      FreeMemory(*Async) ; Parameter freigeben
  EndWith
EndProcedure

; Leseprocedure für Asyncrones lesen
Procedure _SC_Async_Reader(*Async.SC_Async_Launcher)
  Protected *pMem, isUDCached
  
  With *Async
    _SC_LoadEntry_Read(\SC) ; Daten in Current lesen
    If \ParentWriter
      JQ_JobReturnValue(\SC\JQ, \ParentWriter) ; Vorangegangenen Writer vom Stapel nehmen
    EndIf
    
    If isUDCached ; wir sind fertig.
      \SCE\Dyn\AsyncInProgress = #False ; Async Fertig
      FreeMemory(*Async) ; Parameter freigeben
    Else ; Wir müssen erst entpacken
      *Async\Mem = *pMem
      JQ_Add(\SC\JQ, @_SC_Async_UnCompressor(), *Async, #Job_HighPriority)
    EndIf
  EndWith
EndProcedure

; Schreibeprocedure für Asyncrones schreiben
Procedure _SC_Async_Writer(*Async.SC_Async_Launcher)
  With *Async
    _SC_WriteData(\SC, \SCE, \PackedMem) ; Komprimierte Daten in Datei schreiben
    JQ_JobReturnValue(\SC\JQ, \CompressJob) ; Kompressorjob vom Stapel holen
    If \ParentWriter
      JQ_JobReturnValue(\SC\JQ, \ParentWriter) ; Vorangegangenen Writer vom Stapel nehmen
    EndIf
    
    \SCE\Dyn\AsyncInProgress = #False ; Asyncron Fertig :D
    
    FreeMemory(*Async) ; Parameter freigeben
  EndWith
EndProcedure

; Kompressor für Asyncrones schreiben
Procedure _SC_Async_Compressor(*Async.SC_Async_Launcher)
  Protected pMemSize
  
  With *Async
    _SC_CompressData(\SC, \SCE, \PackerInfo, \Mem, \MemSize, \PackedMem, pMemSize) ; Daten komprimieren
    
    If \SC\AsyncCopyMemory ; wurde der Speicher kopiert?
      FreeMemory(\Mem) ; Speicher freigeben
    EndIf
    
    LockMutex(\SC\AsyncAccessMutex) ; Schreibzugriffe auf Datei müssen geordnet werden!
    *Async\ParentWriter = \SC\AsyncLastAccessJob; Parentwriter setzen
    \SC\AsyncLastAccessJob = JQ_Add(\SC\JQ, @_SC_Async_Writer(), *Async, #Job_HighPriority, \SC\AsyncLastAccessJob)
    UnlockMutex(\SC\AsyncAccessMutex)
    
  EndWith
EndProcedure


; IDE Options = PureBasic 4.51 (Windows - x64)
; CursorPosition = 223
; Folding = ---
; EnableXP
; CPU = 1