; ------------------------------------------------------------------------------------
; Dynamischer Cache fr Temporrdaten (angepasst fr ECF (Integer - Indexes, kein MaxAge ... + MinDeadLine!))
; (c) Alexander Aigner
; PB 4.3 + / x86 + x64 / Ascii + Unicode
; Modifikationen fr ECF
; ------------------------------------------------------------------------------------

; ------------------------------------------------------------------------------------
;- Internal
; ------------------------------------------------------------------------------------

XIncludeFile "common\FastMem.pbi"
XIncludeFile "common\LinkedList.pbi"
XIncludeFile "common\SArray.pbi"

; Cache Eintrag
Structure Cache_Entry
  Name.q
  *Mem
  MemSize.i
  Birth.i
  LastAccess.i
  AccessCount.i
  MinDeadline.i
  MemChanged.a
  *CommitEntry
EndStructure

; Cache Daten
Structure Cache_Data
  ShortesDeadLine.i
  *LList.LL_Data
  *CommitList.LL_Data
  SArray.i
  MaxSize.i
  CurrentSize.i
  AgeTolerance.i
  AccessTolerance.i
  ECF.i ; ECF - Anpassung, ECF handle des bergeordneten Objekts
  *Current.Cache_Entry
EndStructure

; aktuellen Entry lschen
Procedure _Cache_DeleteCurrentEntry(*Cache.Cache_Data)
  
  Protected *CEntry.Cache_Entry
  With *Cache
    
    *CEntry = LL_Get(\LList) ; Entry holen
    
    SArray_Set(\SArray, *CEntry\Name, #Null) ; SArray Eintrag Nullsetzen
    \CurrentSize-*CEntry\MemSize ; Speichergre abziehen
    If \Current = *CEntry
      \Current = #Null
    EndIf
    
    If *CEntry\MemChanged And \ECF; ECF - Anpassung!
      ; Cachenderungen auf Datei bertragen, Name = Blockindex
      ECF_WriteBlock(\ECF, *CEntry\Name, *CEntry\Mem, #True)
      If *CEntry\CommitEntry ; Commiteintrag lschen
        LL_SetEntry(\CommitList, *CEntry\CommitEntry)
        LL_Delete(\CommitList)
      EndIf
    EndIf
    
    ; Alles freigeben
    FreeMemory(*CEntry\Mem)
    FreeMemory(*CEntry)
    
    LL_Delete(\LList)
    LL_Reset(\LList)
    
  EndWith
EndProcedure

; Rumt so lange alte Eintrge aus bis genug Platz ist
Procedure _Cache_ClearOld(*Cache.Cache_Data, SizeToAdd)
  Protected *CEntry.Cache_Entry, cEntry, oldest, oldestBirth, cAccessCount, sDeadLine
  With *Cache
    If (SizeToAdd< = *Cache\MaxSize) And *Cache\MaxSize ; Passt das berhaupt rein?
      
      If \ShortesDeadLine ; Uh, es waren nur geschtzte Eintrge im Cache
        If Date() < \ShortesDeadLine
          ProcedureReturn #Null  
        Else ; Aber jetzt kann man mindestens 1. wieder freigeben :D
          \ShortesDeadLine = #Null
        EndIf
      EndIf
      
      While (\MaxSize-\CurrentSize)<SizeToAdd ; Muss was freigeben werden?
        oldest = #Null
        cEntry = #Null
        cAccessCount = #Null
        
        LL_Reset(\LList)
        While LL_Next(\LList) ; Alle abklappern
          
          *CEntry = LL_Get(\LList) ; Eintrag holen

          ; Finden nach: Letzter Zugriff/Anzahl Zugriffe/Wann gechached
          ; Durch die Toleranzen kann "virtuel" ein Eintrag erzeugt werden der etwas schlechter ist als in Wirklichkeit.
          ; z.B. Bei einem Alter in der Toleranz mit weniger Zugriffen wird das schlechtere Alter und die schlechtere Zugriffszeit genommen
          ; Dadurch lsst sich die Auswahl der am wenigsten genutzen Eintrge optimieren
          If *CEntry\MinDeadline >= Date() ; Eintrag kann noch nicht gelscht werden, zu jung :D
            ; Krzeste Deadline finden
            If Not sDeadLine
              sDeadLine = *CEntry\MinDeadline
            ElseIf *CEntry\MinDeadline < sDeadLine
              sDeadLine = *CEntry\MinDeadline
            EndIf
            Continue            
          ElseIf (*CEntry\LastAccess<oldest) Or (oldest = #Null)
            cEntry = LL_GetEntry(\LList) ; ltester unbenutzer Eintrag
            oldest = *CEntry\LastAccess
            oldestBirth = *CEntry\Birth
            cAccessCount = *CEntry\AccessCount
          ElseIf *CEntry\LastAccess-oldest< = \AgeTolerance ; Zugriffszeit in der Toleranz...
            If *CEntry\AccessCount<cAccessCount ;...aber weniger Zugriffe!
              cEntry = LL_GetEntry(\LList)
              cAccessCount = *CEntry\AccessCount
            ElseIf *CEntry\AccessCount-cAccessCount< = \AccessTolerance ; Zugriffe in der Toleranz...
              If *CEntry\Birth<oldestBirth ;...aber ltere Daten!
                cEntry = LL_GetEntry(\LList)
                oldestBirth = *CEntry\Birth
              EndIf
            EndIf
          EndIf
        Wend
        
        If cEntry ; ltestes Lschen
          LL_SetEntry(\LList, cEntry)
          _Cache_DeleteCurrentEntry(*Cache)
        ElseIf (\MaxSize-\CurrentSize)<SizeToAdd ; Uh, konnte nicht genug freigeben
          \ShortesDeadLine = sDeadLine ; Krzeste Deadline bis wieder was geht.
          ProcedureReturn #False
        EndIf
        
      Wend
      
      ProcedureReturn #True
    Else
      ProcedureReturn #True
    EndIf
  EndWith
EndProcedure

Declare Cache_AddEntry(*Cache.Cache_Data, Index.q, *Mem, MemSize, MinLifeTime = 30)

; ------------------------------------------------------------------------------------
;- Public
; ------------------------------------------------------------------------------------

; Cache erzeugen
; MaximumSize: Maximale Gre des Cache in Bytes, fr Unbegrenzt auf #NULL lassen
; AgeTolerance: AltersToleranz bei der beim Speicherfreirumen noch nach schlechteren Daten gesucht wird
; AccessTolerance: ZugriffssToleranz bei der beim Speicherfreirumen noch nach schlechteren Daten gesucht wird
Procedure Cache_New(MaximumSize = #Null, AgeTolerance = #Null, AccessTolerance = #Null)
  Protected *Cache.Cache_Data = AllocateMemory(SizeOf(Cache_Data))
  
  If MaximumSize < #Null : MaximumSize = #Null : EndIf
  
  With *Cache
    \LList = LL_New()
    \CommitList = LL_New()
    \SArray = SArray_New()
    \MaxSize = MaximumSize
    \AgeTolerance = AgeTolerance
    \AccessTolerance = AccessTolerance
    ProcedureReturn *Cache
  EndWith
EndProcedure

; Neu Konfigurieren, Parameter siehe Cache_New()
Procedure Cache_Config(*Cache.Cache_Data, MaximumSize = #Null, AgeTolerance = #Null, AccessTolerance = #Null)
  With *Cache
    \MaxSize = MaximumSize
    \AgeTolerance = AgeTolerance
    \AccessTolerance = AccessTolerance
  EndWith
EndProcedure

; Optimiert den Cache und stellt oft bentigte Daten nach vorne
Procedure Cache_Optimize(*Cache.Cache_Data)
  Protected Change, cAccess, cAccessCount, *CEntry.Cache_Entry
  With *Cache
    ; nach Zugriffszeit+Zugriffen sortieren
    ; (Bubble - Sort)
    Change = #True
    While Change
      Change = #False
      LL_Reset(\LList)
      If LL_Prev(\LList)
        *CEntry = LL_Get(\LList)
        cAccess = *CEntry\LastAccess
        cAccessCount = *CEntry\AccessCount
        While LL_Prev(\LList)
          *CEntry = LL_Get(\LList)
          If *CEntry\LastAccess<cAccess
            LL_SwitchNext(\LList)
            Change = #True
          ElseIf *CEntry\LastAccess = cAccess
            If *CEntry\AccessCount<cAccessCount
              LL_SwitchNext(\LList)
              Change = #True
            EndIf
          EndIf
          cAccess = *CEntry\LastAccess
          cAccessCount = *CEntry\AccessCount
        Wend
      EndIf
    Wend
  EndWith
EndProcedure

; Manuelles lschen des gesamten Cache
Procedure Cache_Flush(*Cache.Cache_Data)
  With *Cache
    LL_Reset(\LList) ; Alle durchgehen
    While LL_Next(\LList)
      _Cache_DeleteCurrentEntry(*Cache)
    Wend
  EndWith
EndProcedure

; schreibt alle genderten Daten in das ECF
Procedure Cache_Commit(*Cache.Cache_Data)
  Protected *CEntry.Cache_Entry
  With *Cache
    If \ECF
      
      LL_Reset(\CommitList) ; Alle durchgehen
      While LL_Next(\CommitList)
        *CEntry = LL_Get(\CommitList) ; Entry holen
        LL_Delete(\CommitList) ; von der Liste lschen
        If *CEntry\MemChanged ; wir haben eine nderung (eigendlich unntig, das Entry landet nur auf der CommitListe wenn es gendert wurde)
          *CEntry\MemChanged = #False
          ; Cachenderungen auf Datei bertragen, Name = Blockindex
          ECF_WriteBlock(\ECF, *CEntry\Name, *CEntry\Mem)
        EndIf
      Wend
      
    EndIf
  EndWith
EndProcedure

; Cache freigeben
Procedure Cache_Free(*Cache.Cache_Data)
  With *Cache
    Cache_Flush(*Cache)
    LL_Free(\LList)
    LL_Free(\CommitList)
    SArray_Free(\SArray)
    FreeMemory(*Cache) ; Cache selbst freigeben
  EndWith
EndProcedure

; Nach gespeicherten Daten suchen
; Name$: Name des Cache Eintrag
Procedure Cache_SetECF(*Cache.Cache_Data, ECF)
  With *Cache
    \ECF = ECF
  EndWith
EndProcedure

; aktuelle Speichergre des Cache
Procedure Cache_GetSize(*Cache.Cache_Data)
  ProcedureReturn *Cache\CurrentSize
EndProcedure

; Nach gespeicherten Daten suchen
; Name$: Name des Cache Eintrag
Procedure Cache_isCached(*Cache.Cache_Data, Index.q)
  Protected *CEntry.Cache_Entry, LLElement
  
  With *Cache
    If \Current
      If Index = \Current\Name
        *CEntry = LL_Get(\LList)
        \Current\AccessCount + 1 ; Zugriffszhler erhhen
        ProcedureReturn #True
      EndIf
    EndIf
    
    LLElement = SArray_Get(\SArray, Index)
    If LLElement 
      LL_SetEntry(\LList, LLElement)
      *CEntry = LL_Get(\LList)
      \Current = *CEntry
      \Current\AccessCount + 1 ; Zugriffszhler erhhen
      *CEntry\LastAccess = Date() ; Letzter Zugriff
      ProcedureReturn #True
    EndIf
    
  EndWith
EndProcedure

; Lscht einen Speicherblock aus dem Cache
; Name$: Name des Cache Eintrag
Procedure Cache_DeleteEntry(*Cache.Cache_Data, Index.q)
  If Cache_isCached(*Cache, Index)
    _Cache_DeleteCurrentEntry(*Cache)
  EndIf
EndProcedure

; Eintrag in den Cache hinzufgen
; Name$: Name des Cache Eintrags, kann spter damit angesprochen werden
; *Mem: Speicherblock
; MemSize: Speicherblockgre
; MaxAge: Maximales Alter des Blocks in Sekunden, fr Unbegrenzt auf #NULL lassen
; XtraData: Extra Infos fr den Cache - Eintrag
; MinLifeTime: Minimale Lebenszeit des Eintrags in Sekunden
Procedure Cache_AddEntry(*Cache.Cache_Data, Index.q, *Mem, MemSize, MinLifeTime = 30)
  Protected *CEntry.Cache_Entry, *tMem
  
  If (MemSize< = *Cache\MaxSize) Or (*Cache\MaxSize = #Null) ; Passt das Rein?
    
    If _Cache_ClearOld(*Cache, MemSize)
      ; Hier sollte man prfen ob der Speicher auch wirklich vorhanden ist
      *tMem = AllocateMemory(MemSize)
      If Not *tMem ; Ui, zu wenig Speicher vorhanden...
        ProcedureReturn #False
      Else ; Kopie erzeugen
        CopyMemory(*Mem, *tMem, MemSize)
      EndIf
    
      ; Cache_DeleteEntry(*Cache, Index) ; die berprfung auf doppelte Indizes muss selbst stattfinden.
    
      *CEntry = AllocateMemory(SizeOf(Cache_Entry))
      With *CEntry
      
        ; Alles eintragen
        \Name = Index
        \Mem = *tMem
        \MemSize = MemSize
        \Birth = Date()
        \LastAccess = \Birth
        \MinDeadline = \Birth + MinLifeTime
      
        ; Und ab auf den Stack
        *Cache\CurrentSize + MemSize ; Cache wchst
        LL_Add(*Cache\LList) : LL_Set(*Cache\LList, *CEntry)
        ; Und ins SArray
        SArray_Set(*Cache\SArray, Index, LL_GetEntry(*Cache\LList))
        
        ProcedureReturn #True
      EndWith
    EndIf
  EndIf
  
EndProcedure

; Speicheraddresse des aktuellen Cache - Eintrags
; Kann nach Cache_isEntry()oder Cache_NextEntry()verwendet werden
Procedure Cache_GetMem(*Cache.Cache_Data)
  With *Cache
    If \Current
      ProcedureReturn \Current\Mem
    EndIf
  EndWith
EndProcedure

; Speichergre des aktuellen Cache - Eintrags
; Kann nach Cache_isEntry()oder Cache_NextEntry()verwendet werden
Procedure Cache_GetMemSize(*Cache.Cache_Data)
  With *Cache
    If \Current
      ProcedureReturn *Cache\Current\MemSize
    EndIf
  EndWith
EndProcedure

; Fr ECF. Der Speicherblock wurde gendert
Procedure Cache_SetChangedFlag(*Cache.Cache_Data)
  With *Cache
    If \Current
      If Not \Current\MemChanged
        \Current\MemChanged = #True
        LL_Add(\CommitList) : LL_Set(\CommitList, \Current) ; Auf Changed-List setzen
        \Current\CommitEntry = LL_GetEntry(\CommitList) ; Eintrag in der Commit-Liste rckspeichern (falls der Eintrag gelscht wird)
      EndIf
    EndIf
  EndWith
EndProcedure
; IDE Options = PureBasic 4.51 (Windows - x64)
; CursorPosition = 13
; FirstLine = 9
; Folding = ---
; EnableXP
; CPU = 1
; DisableDebugger
; EnableCompileCount = 12
; EnableBuildCount = 0
; EnableExeConstant