﻿; EnableExplicit

; aPLib 1.01
; http://www.ibsensoftware.com/products_aPLib.html

; This product uses the aPLib compression library,
; Copyright 1998-2009 by Jørgen Ibsen, All Rights Reserved.
; aPLib is free to use even for commercial use. 

; Wrapper 1.0 for PB 4.41 by Luis (Feb 2010)

; In brief:
; compression: good ratio, average speed, low memory requirements (around 640 KB)
; decompression: extremely fast, extremely low memory requirements 

; all ok
#APLIB_OK               = 0

; error codes for aPLib
#APLIB_ERROR            = 1
#APLIB_ERROR_MEM        = 2
#APLIB_ERROR_OUTPUT     = 3
#APLIB_ERROR_PARAM      = 4

CompilerIf (#PB_Compiler_Processor = #PB_Processor_x86)
 ImportC "common\libs\" + "aPlib_32.lib" 
    aP_pack.i (*SourceBuff, *DestBuff, iSourceSize, *WorkBuff, *callback, *cbparam)
    aP_workmem_size.i (input_size)
    aP_max_packed_size.i (input_size)
    aP_depack_asm_safe.i (*SourceBuff, iSourceSize, *DestBuff, iDestSize)   
 EndImport
CompilerElse   
 ImportC "common\libs\" + "aPlib_64.lib" 
    aP_pack.i (*SourceBuff, *DestBuff, iSourceSize, *WorkBuff, *callback, *cbparam)
    aP_workmem_size.i (input_size)
    aP_max_packed_size.i (input_size)
    aP_depack_asm_safe.i (*SourceBuff, iSourceSize, *DestBuff, iDestSize)   
 EndImport
CompilerEndIf


Procedure.i CalcDestBufferSize_APLIB (iSourceBuffSize)
 ProcedureReturn aP_max_packed_size(iSourceBuffSize)
EndProcedure

 ; *SourceBuff  = ptr to the source buffer
 ; *DestBuff    = ptr to the destination buffer
 ; iSourceSize  = size of the source buffer
 ; *iDestSize   = ptr to the size of the destination buffer (it will contain the packed size on exit)

 ; NOTE: The size of the destination buffer *should* be at least the size returned by CalcDestBufferSize_APLIB()

 ; NOTE: The destination buffer will contain a LONG followed by the packed data
 
 ; NOTE: #APLIB_ERROR_OUTPUT means the file cannot be compressed (or the destination buffer is too small)
Procedure.i PackMemory_APLIB (*SourceBuff, *DestBuff, iSourceSize, *iDestSize.Integer)
 Protected iResult, iCompressedSize, *WorkBuff
  
 *WorkBuff = AllocateMemory(aP_workmem_size(iSourceSize)) ; it should be hard-coded 640 KB
 
;  If *WorkBuff = #Null
;     ProcedureReturn #APLIB_ERROR_MEM
;  EndIf
 
 iCompressedSize = aP_pack(*SourceBuff, *DestBuff, iSourceSize, *WorkBuff, #Null, 0)
 
 If iCompressedSize < 0
    iResult = #APLIB_ERROR
 ElseIf iCompressedSize >= iSourceSize
    iResult = #APLIB_ERROR_OUTPUT    
    ; return -1 for the size of the compressed data 
    PokeI(*iDestSize, -1) ; cannot compress this
 Else              
    
    ; return the size of the compressed data starting at *DestBuff
    PokeI(*iDestSize, iCompressedSize) 
    
    iResult = #APLIB_OK
 EndIf    
 
 FreeMemory(*WorkBuff)
    
 ProcedureReturn iResult
EndProcedure
  
 ; *SourceBuff  = ptr to the source buffer
 ; *DestBuff    = ptr to the destination buffer
 ; iSourceSize  = size of the source buffer
 ; *iDestSize   = ptr to the size of the destination buffer (it will contain the unpacked size on exit)
  
 ; NOTE: The source buffer must contain a LONG followed by the packed data
Procedure.i UnpackMemory_APLIB (*SourceBuff, *DestBuff, iSourceSize, *iDestSize)
 Protected iResult, iDecompressedSize
 Protected iDestSize = PeekI(*iDestSize)

 iDecompressedSize = aP_depack_asm_safe (*SourceBuff, iSourceSize, *DestBuff, iDestSize)
 
 If iDecompressedSize < 0
    iResult = #APLIB_ERROR
 Else
    iResult = #APLIB_OK
    PokeI(*iDestSize, iDecompressedSize)
 EndIf
  
 ProcedureReturn iResult
EndProcedure



; IDE Options = PureBasic 4.51 (Windows - x64)
; CursorPosition = 32
; FirstLine = 7
; Folding = -
; EnableXP
; EnableUser
; Executable = test.exe
; CPU = 1